from sqlalchemy.types import DECIMAL

from . import idsorted
from ..model import Fruit, Tree


def test_get_tree_with_property(rest, http):
    rest(Tree, properties=['fruit_colors'])
    code, json = http.get('/api/tree')
    assert code == 200
    assert json['occurences'] == 3
    assert idsorted(json['objects']) == [
        {'id': 1, 'name': 'pine', 'fruit_colors': 'grey, darkgrey, brown'},
        {'id': 2, 'name': 'maple', 'fruit_colors': 'red, orangered'},
        {'id': 3, 'name': 'oak', 'fruit_colors': ''},
    ]


def test_get_tree_with_object_property(rest, http):
    rest(Tree, properties=[rest.Property('fruit_colors')])
    code, json = http.get('/api/tree')
    assert code == 200
    assert json['occurences'] == 3
    assert idsorted(json['objects']) == [
        {'id': 1, 'name': 'pine', 'fruit_colors': 'grey, darkgrey, brown'},
        {'id': 2, 'name': 'maple', 'fruit_colors': 'red, orangered'},
        {'id': 3, 'name': 'oak', 'fruit_colors': ''},
    ]


results = [
    {'fruit_id': 1, 'color': 'grey', 'size': 12.0,
     'age': 1041300.0, 'tree_id': 1, 'square_size': 144.0},
    {'fruit_id': 2, 'color': 'darkgrey', 'size': 23.0,
     'age': 4233830.213, 'tree_id': 1, 'square_size': 529.0},
    {'fruit_id': 3, 'color': 'brown', 'size': 2.12,
     'age': 0.0, 'tree_id': 1, 'square_size': 4.4944},
    {'fruit_id': 4, 'color': 'red', 'size': 0.5,
     'age': 2400.0, 'tree_id': 2, 'square_size': 0.25},
    {'fruit_id': 5, 'color': 'orangered', 'size': 100.0,
     'age': 7200.000012, 'tree_id': 2, 'square_size': 10000.0}]


def test_get_fruit_with_hybrid_property_and_type(rest, http):
    rest(Fruit, properties=[rest.Property('square_size', DECIMAL())])
    code, json = http.get('/api/fruit')
    assert code == 200
    assert json['occurences'] == 5
    assert idsorted(json['objects'], 'fruit_id') == results


def test_get_fruit_with_hybrid_property_and_formatter(rest, http):
    rest(Fruit, properties=[rest.Property(
        'square_size', formatter=lambda x: float(x))])
    code, json = http.get('/api/fruit')
    assert code == 200
    assert json['occurences'] == 5
    assert idsorted(json['objects'], 'fruit_id') == results



from unrest import UnRest
from unrest.flask_framework import FlaskUnRest

from ..model import Fruit, Tree


def test_normal(app, db, http):
    rest = UnRest(app, db.session)
    rest(Tree)
    code, json = http.get('/api/tree')
    assert code == 200
    assert json['occurences'] == 3


def test_path(app, db, http):
    rest = UnRest(app, db.session, '/forest')
    rest(Tree)
    code, json = http.get('/forest/tree')
    assert code == 200
    assert json['occurences'] == 3


def test_version(app, db, http):
    rest = UnRest(app, db.session, version='v3.14')
    rest(Tree)
    code, json = http.get('/api/v3.14/tree')
    assert code == 200
    assert json['occurences'] == 3


def test_path_and_version(app, db, http):
    rest = UnRest(app, db.session, '/forest', 'v3.14')
    rest(Tree)
    code, json = http.get('/forest/v3.14/tree')
    assert code == 200
    assert json['occurences'] == 3


def test_explicit_framework(app, db, http):
    rest = UnRest(app, db.session, framework=FlaskUnRest)
    rest(Tree)
    code, json = http.get('/api/tree')
    assert code == 200
    assert json['occurences'] == 3


def test_empty_explicit_framework(app, db, http):

    class FakeUnRest(object):
        def __init__(self, app):
            pass

        def register_route(self, *args, **kwargs):
            pass
        request_json = register_route
        send_json = register_route
        send_error = register_route

    rest = UnRest(app, db.session, framework=FakeUnRest)
    rest(Tree)
    code, _ = http.get('/api/tree')
    assert code == 404


def test_api_options(app, db, http):
    rest = UnRest(app, db.session, framework=FlaskUnRest)
    fruit = rest(Fruit)
    rest(Tree, methods=rest.all,
         relationships={'fruits': fruit},
         properties=['fruit_colors'],
         allow_batch=True)
    code, json = http.options('/api')
    assert code == 200
    assert json == {
        '/api/fruit': {
            'model': 'Fruit',
            'description': 'A bag of fruit',
            'columns': {
                'age': 'timedelta',
                'color': 'str',
                'fruit_id': 'int',
                'size': 'Decimal',
                'tree_id': 'int'
            },
            'methods': ['GET', 'OPTIONS']
        },
        '/api/tree': {
            'model': 'Tree',
            'description': "Where money doesn't grow",
            'columns': {
                'id': 'int',
                'name': 'str'
            },
            'properties': {
                'fruit_colors': 'The color of fruits'
            },
            'relationships': {
                'fruits': {
                    'model': 'Fruit',
                    'description': 'A bag of fruit',
                    'columns': {
                        'age': 'timedelta',
                        'color': 'str',
                        'fruit_id': 'int',
                        'size': 'Decimal',
                        'tree_id': 'int'
                    }
                }
            },
            'methods': ['GET', 'PUT', 'POST', 'DELETE', 'PATCH', 'OPTIONS'],
            'batch': True
        }
    }


def test_endpoint_options(app, db, http):
    rest = UnRest(app, db.session, framework=FlaskUnRest)
    fruit = rest(Fruit)
    rest(Tree, methods=rest.all,
         relationships={'fruits': fruit},
         properties=['fruit_colors'],
         allow_batch=True)
    code, json = http.options('/api/fruit')
    assert code == 200
    assert json == {
        'model': 'Fruit',
        'description': 'A bag of fruit',
        'columns': {
            'age': 'timedelta',
            'color': 'str',
            'fruit_id': 'int',
            'size': 'Decimal',
            'tree_id': 'int'
        },
        'methods': ['GET', 'OPTIONS']
    }



import json
import logging
from functools import wraps

from sqlalchemy import and_, or_
from sqlalchemy.inspection import inspect
from sqlalchemy.schema import Column

from .coercers import Deserialize, Serialize

try:
    from json import JSONDecodeError
except ImportError:
    JSONDecodeError = Exception


log = logging.getLogger('unrest.rest')


class Rest(object):
    """
    This is the entry point for generating a REST endpoint for a specific model
    The final uri if the path is '/api' and version 'v2' would be:
    `/api/v2/model` and `/api/v2/model/pk1/pk2` and if model is not in the
    public schema `/api/v2/schema/model` and `/api/v2/schema/model/pk1/pk2`.

    Usage:
    ```python
        rest = UnRest(app)
        rest(Person, only=['name', 'sex', 'age'], methods=rest.all,
             query=lambda q: q.filter(Person.age > 16))
    ```
    # Arguments
        unrest: The unrest instance given automatically on UnRest call.
        Model: The sqlalchemy orm model class.
        methods: The allowed method list on this endpoint. Possible values are
            GET, PUT, POST, DELETE, PATCH and rest.all
        name: If specified replaces the model name in url.
        only: If specified restricts the json fields to this list.
        exclude: If specified removes the json fields in this list.
        query: A function that takes the Model query and returns your specific
            query. Can be useful to filter data for all the methods.
        properties: A list of additional properties to retrieve on the model.
        relationships: A mapping of relationships and rest endpoints to fetch
            with the model.
        allow_batch: Allow batch operations (PUT, DELETE and PATCH)
            without primary key.
        auth: A decorator that will always be called.
        read_auth: A decorator that will be called on GET.
        write_auth: A decorator that will be called on PUT, POST, DELETE
            and PATCH.
        SerializeClass: An alternative #Serialize class.
        DeserializeClass: An alternative #Deserialize class.
    """
    def __init__(self, unrest, Model,
                 methods=['GET'], name=None, only=None, exclude=None,
                 query=None, properties=None, relationships=None,
                 allow_batch=False, auth=None, read_auth=None, write_auth=None,
                 SerializeClass=Serialize, DeserializeClass=Deserialize):
        self.unrest = unrest
        self.Model = Model

        self.methods = methods[:]
        self.name = name or self.table.name
        self.only = only
        self.exclude = exclude
        self.query_factory = query or (lambda q: q)
        self.properties = [
            self.unrest.Property(property)
            if not isinstance(property, self.unrest.Property) else property
            for property in (properties or [])
        ]
        self.relationships = relationships or {}

        self.allow_batch = allow_batch

        self.auth = auth
        self.read_auth = read_auth
        self.write_auth = write_auth

        self.SerializeClass = SerializeClass
        self.DeserializeClass = DeserializeClass

        self.infos = self.unrest.infos[self.path]

        self.set_infos()
        if self.unrest.allow_options and self.methods:
            self.methods.append('OPTIONS')

        for method in self.methods:
            self.register_method(method)

    def get(self, payload, **pks):
        """
        The GET method

        No arguments: Returns all query elements. (/api/model/)
        Primary keys: Returns the element in query with the primary keys or
            404. (/api/model/pk)

        # Arguments
            payload: The json request content ignored for GET.
            pks: The primary keys in url if any.
        """
        if self.has(pks):
            item = self.get_from_pk(self.query, **pks)
            if item is None:
                self.raise_error(404, '%s(%r) not found' % (self.name, pks))

            return self.serialize([item])

        items = self.query
        return self.serialize(items)

    def put(self, payload, **pks):
        """
        The PUT method

        No arguments: If allow_batch set to true replace all the query elements
            with the ones in the request payload.
        Primary keys: Create or replace the element associated
            with the primary keys from the one in the request payload.

        # Arguments
            payload: The json request content containing new elements.
            pks: The primary keys in url if any.
        """
        if not payload:
            self.raise_error(400, 'You must provide a payload')

        if self.has(pks):
            for pk, val in pks.items():
                if pk in payload:
                    assert payload[pk] == val, (
                        'Incoherent primary_key (%s) in payload (%r) '
                        'and url (%r) for PUT' % (pk, payload[pk], val))
                else:
                    payload[pk] = val
            existingItem = self.get_from_pk(self.query, **pks)
            item = self.deserialize(payload, existingItem or self.Model())
            if existingItem is None:
                self.session.add(item)
            self.session.commit()
            return self.serialize([item])

        if not self.allow_batch:
            raise self.unrest.RestError(
                406, 'You must set allow_batch to True '
                'if you want to use batch methods.')

        self.query.delete()
        items = self.deserialize_all(payload)
        self.session.add_all(items)
        self.session.commit()
        return self.serialize(items)

    def post(self, payload, **pks):
        """
        The POST method

        No arguments: Add element from request payload.
        Primary keys: Correspond to new collection creation. Unused.

        # Arguments
            payload: The json request content containing the new element.
            pks: The primary keys in url if any.
        """
        if self.has(pks):
            # Create a collection?
            raise self.unrest.RestError(
                501, "POST with primary keys corresponds to collection "
                "creation. It's not implemented by default. "
                "If you want to update an item use the PUT method instead")

        if not payload:
            self.raise_error(400, 'You must provide a payload')
        item = self.deserialize(payload, self.Model())
        self.session.add(item)
        self.session.commit()
        return self.serialize([item])

    def delete(self, payload, **pks):
        """
        The DELETE method

        No arguments: If allow_batch set to true delete all query elements.
        Primary keys: Delete the element associated with the primary keys.

        # Arguments
            payload: The json request content ignored in DELETE.
            pks: The primary keys of the element to delete.
        """
        if self.has(pks):
            item = self.get_from_pk(self.query, **pks)
            if item is None:
                self.raise_error(404, '%s(%r) not found' % (self.name, pks))

            self.session.delete(item)
            self.session.commit()
            return self.serialize([item])

        if not self.allow_batch:
            raise self.unrest.RestError(
                406, 'You must set allow_batch to True '
                'if you want to use batch methods.')

        items = self.query.all()
        self.query.delete()
        self.session.commit()
        return self.serialize(items)

    def patch(self, payload, **pks):
        """
        The PATCH method

        No arguments: If allow_batch set to true patch existing elements
            with element attributes specified in the request payload.
        Primary keys: Patch only one

        # Arguments
            payload: The json request content containing
                a list of attributes to be patched.
            pks: The primary keys of the element to patch.
        """
        if not payload:
            self.raise_error(400, 'You must provide a payload')

        if self.has(pks):
            for pk, val in pks.items():
                if pk in payload:
                    assert payload[pk] == val, (
                        'Incoherent primary_key (%s) in payload (%r) '
                        'and url (%r) for PATCH' % (pk, payload[pk], val))
                else:
                    payload[pk] = val
            item = self.get_from_pk(self.query, **pks)
            if item is None:
                self.raise_error(404, '%s(%r) not found' % (self.name, pks))
            self.deserialize(payload, item, blank_missing=False)
            self.session.commit()
            return self.serialize([item])

        if self.has(pks):
            raise self.unrest.RestError(
                501, "PATCH with primary keys corresponds to nothing. "
                "It's not implemented by default. ")

        if not self.allow_batch:
            raise self.unrest.RestError(
                406, 'You must set allow_batch to True '
                'if you want to use batch methods.')

        patches = payload['objects']
        # Get all concerned items
        items = self.get_all_from_pks(self.query, [{
            pk: patch[pk] for pk in self.primary_keys}
            for patch in patches])
        if len(items) < len(patches):
            for patch in patches:
                if len([item for pk in self.primary_keys for item in items
                        if getattr(item, pk) == patch[pk]]) == 0:
                    self.raise_error(404, '%s(%r) not found' % (
                        self.name, {key: val for key, val in patch.items()
                                    if key in self.primary_keys}))
        for patch in patches:
            # Get the patch item
            item = [item
                    for pk in self.primary_keys
                    for item in items if getattr(item, pk) == patch[pk]][0]
            # Merge only patched colmuns
            self.deserialize(patch, item, blank_missing=False)

        self.session.commit()
        return self.serialize(items)

    def options(self, payload):
        """
        The OPTIONS method

        Returns a description of this rest endpoint.
        """
        return self.infos

    def declare(self, method):
        """
        A decorator to register an alternative method.
        The original is still callable with rest.{method}

        ```python
        fruit = rest(Fruit)

        @fruit.declare('GET')
        def get(payload, fruit_id=None):
            rv = fruit.get(payload, fruit_id=fruit_id)
            return {
                'occurences': rv['occurences'],
                'objects': [
                    {'id': obj['fruit_id']} for obj in rv['objects']
                ]
            }
        ```

        # Arguments
            method: The method to override ('GET' for exemple)
        """
        def register_fun(fun):
            if self.unrest.allow_options and not self.methods:
                self.register_method('OPTIONS')
            self.register_method(method, fun)
        return register_fun

    def kwargs_to_pks(self, kwargs):
        if not kwargs:
            return {}
        item = self.Model()
        self.DeserializeClass(kwargs, self.primary_keys).merge(item)
        return {name: getattr(item, name) for name in self.primary_keys}

    def deserialize(self, payload, item, blank_missing=True):
        if blank_missing:
            columns = self.columns
        else:
            # Mind only provided columns
            columns = {
                name: column for name, column in self.columns.items()
                if name in payload}
        return self.DeserializeClass(payload, columns).merge(item)

    def deserialize_all(self, payload):
        return self.DeserializeClass(payload, self.columns).create(self.Model)

    def serialize_object(self, item):
        return self.SerializeClass(
            item, self.columns, self.properties, self.relationships).dict()

    def serialize(self, items):
        objects = [
            self.serialize_object(item) for item in items  # Pagination?
        ]

        return {
            'occurences': len(objects),
            'primary_keys': list(self.primary_keys.keys()),
            'objects': objects
        }

    def raise_error(self, status, message):
        self.unrest.raise_error(status, message)

    def wrap_native(self, method, method_fun):
        @wraps(method_fun)
        def wrapped(**kwargs):
            pks = self.kwargs_to_pks(kwargs)
            json = self.unrest.framework.request_json()
            try:
                try:
                    payload = self.unjson(json)
                except JSONDecodeError as e:
                    self.raise_error(400, 'JSON Error in payload: %s' % e)
                decorated = method_fun
                if method == 'GET' and self.read_auth:
                    decorated = self.read_auth(decorated)
                if (method in ['PUT', 'POST', 'DELETE', 'PATCH'] and
                        self.write_auth):
                    decorated = self.write_auth(decorated)
                if self.auth:
                    decorated = self.auth(decorated)

                data = decorated(payload, **pks)
            except self.unrest.RestError as e:
                return self.unrest.framework.send_error(
                    dict(message=e.message, **e.extra), e.status)
            json = self.json(data)
            return self.unrest.framework.send_json(json)
        return wrapped

    def register_method(self, method, method_fun=None):
        if method != 'OPTIONS':
            assert method in self.unrest.all, 'Unknown method %s' % method
        method_fun = method_fun or getattr(self, method.lower())
        method_fun = self.wrap_native(method, method_fun)
        # str() for python 2 compat
        method_fun.__name__ = str('_'.join(
            ('unrest', method) + self.name_parts))
        self.unrest.framework.register_route(
            self.path, method,
            self.primary_keys if method != 'OPTIONS' else None, method_fun)

        self.infos['methods'].append(method)

    def json(self, data):
        return json.dumps(data)

    def unjson(self, data):
        if data:
            return json.loads(data)

    def has(self, pks):
        return pks and all(val is not None for val in pks.values())

    def set_infos(self):
        self.infos['model'] = self.Model.__name__
        if getattr(self.Model, '__doc__', None):
            self.infos['description'] = self.Model.__doc__

        def sqlatype(type):
            try:
                return type.python_type.__name__
            except NotImplementedError:
                return type.__class__.__name__

        self.infos['columns'] = {
            name: sqlatype(column.type)
            for name, column in self.columns.items()
        }

        if self.properties:
            self.infos['properties'] = {
                prop.name: getattr(
                    getattr(self.Model, prop.name), '__doc__', 'Undocumented')
                for prop in self.properties
            }

        if self.relationships:
            self.infos['relationships'] = {
                rel: {k: v for k, v in rest.infos.items() if k != 'methods'}
                for rel, rest in self.relationships.items()
            }

        if self.allow_batch:
            self.infos['batch'] = self.allow_batch

        self.infos['methods'] = []

    def get_from_pk(self, query, **pks):
        for key, val in pks.items():
            query = query.filter(getattr(self.Model, key) == val)
        return query.first()

    def get_all_from_pks(self, query, items_pks):
        return query.filter(
            or_(*[and_(*[
                getattr(self.Model, key) == val
                for key, val in pks.items()
            ]) for pks in items_pks])).all()

    @property
    def session(self):
        return self.unrest.session

    @property
    def query(self):
        return self.query_factory(self.session.query(self.Model))

    @property
    def name_parts(self):
        if self.table.schema:
            return (self.table.schema, self.name)
        return (self.name,)

    @property
    def path(self):
        return '/'.join((self.unrest.root_path,) + self.name_parts)

    @property
    def table(self):
        return self.Model.__table__

    @property
    def primary_keys(self):
        ins = inspect(self.Model)
        return {ins.get_property_by_column(pk).key: pk
                for pk in ins.primary_key}

    @property
    def model_columns(self):
        for name, column in inspect(self.Model).columns.items():
            if isinstance(column, Column):
                yield name, column

    @property
    def columns(self):
        def gen():
            for name, column in self.model_columns:
                if name not in self.primary_keys:
                    if self.only is not None and name not in self.only:
                        continue
                    if self.exclude and name in self.exclude:
                        continue
                yield name, column
        return dict(gen())



from . import idsorted
from ..model import Fruit, Tree


def test_get_pk_tree(rest, http):
    rest(Tree)
    rest(Fruit)
    code, json = http.get('/api/tree/1')
    assert code == 200
    assert json['occurences'] == 1
    assert idsorted(json['objects']) == [
        {'id': 1, 'name': 'pine'}
    ]


def test_get_pk_unknown_tree(rest, http):
    rest(Tree)
    rest(Fruit)
    code, json = http.get('/api/tree/6')
    assert code == 404
    assert json['message'] == "tree({'id': 6}) not found"


def test_get_pk_tree_name(rest, http):
    rest(Tree, name='forest')
    code, json = http.get('/api/forest/2')
    assert code == 200
    assert json['occurences'] == 1
    assert idsorted(json['objects']) == [
        {'id': 2, 'name': 'maple'},
    ]


def test_get_pk_tree_query(rest, http, db):
    def base_query(q):
        return db.session.query(Tree).filter(Tree.id > 1)
    rest(Tree, query=base_query)
    code, json = http.get('/api/tree/2')
    assert code == 200
    assert json['occurences'] == 1
    assert idsorted(json['objects']) == [
        {'id': 2, 'name': 'maple'},
    ]


def test_get_pk_tree_query_not_found(rest, http, db):
    def base_query(q):
        return db.session.query(Tree).filter(Tree.id > 1)
    rest(Tree, query=base_query)
    code, json = http.get('/api/tree/1')
    assert code == 404
    assert json['message'] == "tree({'id': 1}) not found"


def test_get_pk_tree_query_factory(rest, http, db):
    rest(Tree, query=lambda q: q.filter(Tree.id < 2))
    code, json = http.get('/api/tree')
    assert code == 200
    assert json['occurences'] == 1
    assert idsorted(json['objects']) == [
        {'id': 1, 'name': 'pine'}
    ]


def test_get_pk_tree_query_factory_not_found(rest, http, db):
    rest(Tree, query=lambda q: q.filter(Tree.id < 2))
    code, json = http.get('/api/tree/3')
    assert code == 404
    assert json['message'] == "tree({'id': 3}) not found"


def test_get_pk_fruits(rest, http):
    rest(Tree)
    rest(Fruit)
    code, json = http.get('/api/fruit/1')
    assert code == 200
    assert json['occurences'] == 1
    assert idsorted(json['objects'], 'fruit_id') == [
        {
            'fruit_id': 1,
            'color': 'grey',
            'size': 12.0,
            'age': 1041300.0,
            'tree_id': 1
        }
    ]


def test_get_pk_fruits_only(rest, http):
    rest(Fruit, only=['color', 'size'])
    code, json = http.get('/api/fruit/2')
    assert code == 200
    assert json['occurences'] == 1
    assert idsorted(json['objects'], 'fruit_id') == [
        {
            'fruit_id': 2,
            'color': 'darkgrey',
            'size': 23.0
        }
    ]


def test_get_pk_fruits_exclude(rest, http):
    rest(Fruit, exclude=['color', 'age'])
    code, json = http.get('/api/fruit/3')
    assert code == 200
    assert json['occurences'] == 1
    assert idsorted(json['objects'], 'fruit_id') == [
        {
            'fruit_id': 3,
            'size': 2.12,
            'tree_id': 1
        }
    ]


def test_get_pk_fruits_only_exclude(rest, http):
    rest(Fruit, only=['color', 'size'], exclude=['color', 'age'])
    code, json = http.get('/api/fruit/4')
    assert code == 200
    assert json['occurences'] == 1
    assert idsorted(json['objects'], 'fruit_id') == [
        {
            'fruit_id': 4,
            'size': 0.5
        }
    ]


def test_no_method(rest, http):
    rest(Fruit, methods=[])
    code, json = http.get('/api/fruit/1')
    assert code == 404


def test_get_custom(rest, http):
    fruit = rest(Fruit)

    @fruit.declare('GET')
    def get(payload, fruit_id=None):
        return {'Hey': 'Overridden'}

    code, json = http.get('/api/fruit/3')
    assert code == 200
    assert json == {'Hey': 'Overridden'}


def test_get_custom_extend(rest, http):
    fruit = rest(Fruit)

    @fruit.declare('GET')
    def get(payload, fruit_id=None):
        fruit_id += 1
        return fruit.get(payload, fruit_id=fruit_id)

    code, json = http.get('/api/fruit/4')
    assert code == 200
    assert json['occurences'] == 1
    assert idsorted(json['objects'], 'fruit_id') == [
        {
            'fruit_id': 5,
            'color': 'orangered',
            'size': 100.0,
            'age': 7200.000012,
            'tree_id': 2
        }
    ]



from . import idsorted
from ..model import Tree, Fruit


def test_put_tree(rest, http):
    rest(Tree, methods=['GET', 'PUT'])
    code, json = http.put('/api/tree/1', json={'id': 1, 'name': 'cedar'})
    assert code == 200
    assert json['occurences'] == 1
    assert idsorted(json['objects']) == [
        {'id': 1, 'name': 'cedar'},
    ]

    code, json = http.get('/api/tree')
    assert code == 200
    assert json['occurences'] == 3
    assert idsorted(json['objects']) == [
        {'id': 1, 'name': 'cedar'},
        {'id': 2, 'name': 'maple'},
        {'id': 3, 'name': 'oak'},
    ]


def test_put_tree_without_id(rest, http):
    rest(Tree, methods=['GET', 'PUT'])
    code, json = http.put('/api/tree/1', json={'name': 'cedar'})
    assert code == 200
    assert json['occurences'] == 1
    assert idsorted(json['objects']) == [
        {'id': 1, 'name': 'cedar'},
    ]

    code, json = http.get('/api/tree')
    assert code == 200
    assert json['occurences'] == 3
    assert idsorted(json['objects']) == [
        {'id': 1, 'name': 'cedar'},
        {'id': 2, 'name': 'maple'},
        {'id': 3, 'name': 'oak'},
    ]


def test_put_tree_with_another_id(rest, http):
    rest(Tree, methods=['GET', 'PUT'])
    code, json = http.put('/api/tree/1', json={'id': 5, 'name': 'cedar'})
    assert code == 500


def test_put_fruit(rest, http):
    rest(Fruit, methods=['GET', 'PUT'], allow_batch=True)
    code, json = http.put('/api/fruit/1', json={
        'fruit_id': 1, 'color': 'blue'
    })
    assert code == 200
    assert json['occurences'] == 1
    assert idsorted(json['objects'], 'fruit_id') == [
        {
            'fruit_id': 1,
            'color': 'blue',
            'size': None,
            'age': None,
            'tree_id': None
        }
    ]

    code, json = http.get('/api/fruit')
    assert code == 200
    assert json['occurences'] == 5
    assert idsorted(json['objects'], 'fruit_id') == [
        {
            'fruit_id': 1,
            'color': 'blue',
            'size': None,
            'age': None,
            'tree_id': None
        }, {
            'fruit_id': 2,
            'color': 'darkgrey',
            'size': 23.0,
            'age': 4233830.213,
            'tree_id': 1
        }, {
            'fruit_id': 3,
            'color': 'brown',
            'size': 2.12,
            'age': 0.0,
            'tree_id': 1
        }, {
            'fruit_id': 4,
            'color': 'red',
            'size': 0.5,
            'age': 2400.0,
            'tree_id': 2
        }, {
            'fruit_id': 5,
            'color': 'orangered',
            'size': 100.0,
            'age': 7200.000012,
            'tree_id': 2
        }
    ]



from . import idsorted
from ..model import Tree


def test_delete_tree_implicitly_unallowed(rest, http):
    rest(Tree, methods=['GET', 'DELETE'])
    code, json = http.delete('/api/tree')
    assert code == 406


def test_delete_tree(rest, http):
    rest(Tree, methods=['GET', 'DELETE'], allow_batch=True)
    code, json = http.delete('/api/tree')
    assert code == 200
    assert json['occurences'] == 3
    assert idsorted(json['objects']) == [
        {'id': 1, 'name': 'pine'},
        {'id': 2, 'name': 'maple'},
        {'id': 3, 'name': 'oak'},
    ]

    code, json = http.get('/api/tree')
    assert code == 200
    assert json['occurences'] == 0
    assert idsorted(json['objects']) == []



from . import idsorted
from ..model import Tree


def test_post_tree(rest, http):
    rest(Tree, methods=['GET', 'POST'])
    code, json = http.post('/api/tree', json={'name': 'cedar'})
    assert code == 200
    assert json['occurences'] == 1
    assert idsorted(json['objects']) == [
        {'id': 4, 'name': 'cedar'},
    ]

    code, json = http.get('/api/tree')
    assert code == 200
    assert json['occurences'] == 4
    assert idsorted(json['objects']) == [
        {'id': 1, 'name': 'pine'},
        {'id': 2, 'name': 'maple'},
        {'id': 3, 'name': 'oak'},
        {'id': 4, 'name': 'cedar'},
    ]


def test_post_tree_with_id(rest, http):
    rest(Tree, methods=['GET', 'POST'])
    code, json = http.post('/api/tree', json={'id': 9, 'name': 'cedar'})
    assert code == 200
    assert json['occurences'] == 1
    assert idsorted(json['objects']) == [
        {'id': 9, 'name': 'cedar'},
    ]

    code, json = http.get('/api/tree')
    assert code == 200
    assert json['occurences'] == 4
    assert idsorted(json['objects']) == [
        {'id': 1, 'name': 'pine'},
        {'id': 2, 'name': 'maple'},
        {'id': 3, 'name': 'oak'},
        {'id': 9, 'name': 'cedar'},
    ]


def test_post_tree_custom(rest, http):
    tree = rest(Tree)

    @tree.declare('POST')
    def post(payload, id=None):
        payload['name'] = 'I ALWAYS WANT THIS NAME'
        return tree.post(payload, id=id)

    code, json = http.post('/api/tree', json={'name': 'cedar'})
    assert code == 200
    assert json['occurences'] == 1
    assert idsorted(json['objects']) == [
        {'id': 4, 'name': 'I ALWAYS WANT THIS NAME'},
    ]

    code, json = http.get('/api/tree')
    assert code == 200
    assert json['occurences'] == 4
    assert idsorted(json['objects']) == [
        {'id': 1, 'name': 'pine'},
        {'id': 2, 'name': 'maple'},
        {'id': 3, 'name': 'oak'},
        {'id': 4, 'name': 'I ALWAYS WANT THIS NAME'},
    ]



from datetime import timedelta

from . import idsorted
from ..model import Fruit, Tree


def test_patch_tree_implicitly_unallowed(rest, http):
    rest(Tree, methods=['GET', 'PATCH'])
    code, json = http.patch('/api/tree', json={'id': 1, 'name': 'cedar'})
    assert code == 406


def test_patch_tree(rest, http):
    rest(Tree, methods=['GET', 'PATCH'], allow_batch=True)
    code, json = http.patch('/api/tree', json={
        'objects': [
            {'id': 1, 'name': 'cedar'}, {'id': 2, 'name': 'mango'}
        ]})
    assert code == 200
    assert json['occurences'] == 2
    assert idsorted(json['objects']) == [
        {'id': 1, 'name': 'cedar'},
        {'id': 2, 'name': 'mango'},
    ]

    code, json = http.get('/api/tree')
    assert code == 200
    assert json['occurences'] == 3
    assert idsorted(json['objects']) == [
        {'id': 1, 'name': 'cedar'},
        {'id': 2, 'name': 'mango'},
        {'id': 3, 'name': 'oak'},
    ]


def test_patch_missing_tree(rest, http):
    rest(Tree, methods=['GET', 'PATCH'], allow_batch=True)
    code, json = http.patch('/api/tree', json={
        'objects': [
            {'id': 1, 'name': 'cedar'}, {'id': 8, 'name': 'mango'}
        ]})
    assert code == 404


def test_patch_fruit(rest, http):
    rest(Fruit, methods=['GET', 'PATCH'], allow_batch=True)
    code, json = http.patch('/api/fruit', json={
        'objects': [
            {'fruit_id': 1, 'color': 'blue'},
            {'fruit_id': 3, 'age': timedelta(days=12, minutes=29)
             .total_seconds()},
            {'fruit_id': 4, 'color': 'rainbow', 'size': 8},
            {'fruit_id': 5, 'size': 10, 'tree_id': 1},
        ]})
    assert code == 200
    assert json['occurences'] == 4
    assert idsorted(json['objects'], 'fruit_id') == [
        {
            'fruit_id': 1,
            'color': 'blue',
            'size': 12.0,
            'age': 1041300.0,
            'tree_id': 1
        }, {
            'fruit_id': 3,
            'color': 'brown',
            'size': 2.12,
            'age': 1038540.0,
            'tree_id': 1
        }, {
            'fruit_id': 4,
            'color': 'rainbow',
            'size': 8.0,
            'age': 2400.0,
            'tree_id': 2
        }, {
            'fruit_id': 5,
            'color': 'orangered',
            'size': 10.0,
            'age': 7200.000012,
            'tree_id': 1
        }
    ]

    code, json = http.get('/api/fruit')
    assert code == 200
    assert json['occurences'] == 5
    assert idsorted(json['objects'], 'fruit_id') == [
        {
            'fruit_id': 1,
            'color': 'blue',
            'size': 12.0,
            'age': 1041300.0,
            'tree_id': 1
        }, {
            'fruit_id': 2,
            'color': 'darkgrey',
            'size': 23.0,
            'age': 4233830.213,
            'tree_id': 1
        }, {
            'fruit_id': 3,
            'color': 'brown',
            'size': 2.12,
            'age': 1038540.0,
            'tree_id': 1
        }, {
            'fruit_id': 4,
            'color': 'rainbow',
            'size': 8.0,
            'age': 2400.0,
            'tree_id': 2
        }, {
            'fruit_id': 5,
            'color': 'orangered',
            'size': 10.0,
            'age': 7200.000012,
            'tree_id': 1
        }
    ]



from . import idsorted
from ..model import Tree


def test_put_tree_implicitly_unallowed(rest, http):
    rest(Tree, methods=['GET', 'PUT'])
    code, json = http.put('/api/tree', json={'id': 1, 'name': 'cedar'})
    assert code == 406


def test_put_tree(rest, http):
    rest(Tree, methods=['GET', 'PUT'], allow_batch=True)
    code, json = http.put('/api/tree', json={
        'objects': [
            {'id': 1, 'name': 'cedar'}, {'id': 2, 'name': 'mango'}
        ]})
    assert code == 200
    assert json['occurences'] == 2
    assert idsorted(json['objects']) == [
        {'id': 1, 'name': 'cedar'},
        {'id': 2, 'name': 'mango'},
    ]

    code, json = http.get('/api/tree')
    assert code == 200
    assert json['occurences'] == 2
    assert idsorted(json['objects']) == [
        {'id': 1, 'name': 'cedar'},
        {'id': 2, 'name': 'mango'},
    ]



import json
import logging
from collections import defaultdict

from .coercers import Property
from .rest import Rest


log = logging.getLogger('unrest')


class UnRest(object):
    """
    A troubling rest api library for sqlalchemy models.
    This is the main entry point of unrest.

    Common usage is as following:

    ```python
    rest = UnRest(app, session)  # app is your application
    # When called, it instanciate a `Rest` object that will register the
    # REST endpoints. See the `Rest` class.
    rest(Model1)
    rest(Model2)
    ```

    # Arguments
        app: Your web application,
            can be set afterwards using #UnRest.init_app
        session: Your sqlalchemy session,
            can be set afterwards using #UnRest.init_session
            and `init_session` method.
        path: Default '/api', sets the root url path for your endpoints
        version: Adds a version to the root url path if specified
            (i.e. /api/v2)
        framework: Your specific framework class, defaults to auto detect.
        SerializeClass: A global alternative for #Serialize class.
        DeserializeClass: A global alternative for #Deserialize class.
        allow_options: Set it to False to disable OPTIONS requests
    Unrest aims to be framework agnostic.
    It currently works with Flask out of the box, for another web framework
    you will have to implement your own Framework class.
    See `FlaskUnRest` in `flask_framework.py`
    """

    class RestError(Exception):
        """
        Exception raised by rest methods. It's catched by the REST method
        wrapper and will return a `status` http error with the specified
        `message`.
        """
        def __init__(self, status, message, extra=None):
            self.status = status
            self.message = message
            self.extra = extra or {}

    def __init__(self,
                 app=None, session=None,
                 path='/api', version='', framework=None,
                 SerializeClass=None, DeserializeClass=None,
                 allow_options=True):
        self.path = path
        self.version = version
        self._framework = framework
        self.SerializeClass = SerializeClass
        self.DeserializeClass = DeserializeClass
        self.allow_options = allow_options

        self.infos = defaultdict(dict)

        if app is not None:
            self.init_app(app)
        if session is not None:
            self.init_session(session)

    def init_app(self, app):
        """Sets the app on UnRest if it was missing during instantiation."""
        self.app = app
        if self._framework:
            self.framework = self._framework(app)
        else:
            try:
                from flask import Flask
            except ImportError:
                pass
            else:
                if isinstance(app, Flask):
                    from .flask_framework import FlaskUnRest
                    self.framework = FlaskUnRest(app)
        if not self.framework:
            raise NotImplementedError(
                'Your framework %s is not recognized. '
                'Please provide a framework argument to UnRest' % type(app))
        self.allow_options and self.register_options()

    def init_session(self, session):
        """
        Sets the sqlalchemy session on UnRest
        if it was missing during instantiation.
        """
        self.session = session

    @property
    def root_path(self):
        if self.version:
            return '/'.join((self.path, self.version))
        return self.path

    @property
    def all(self):
        """
        Return all supported methods. Useful for the rest
        `method` keyword argument.
        """
        return ['GET', 'PUT', 'POST', 'DELETE', 'PATCH']

    def raise_error(self, status, message):
        """
        Raise an error that will be handled by the rest wrapper, which
        will return a json response with status as HTTP status code
        and message as content.

        # Arguments
            status: The http status code corresponding to the error
                (404 for instance)
            message: The message that will be returned in the json response
        """
        raise self.RestError(status, message)

    def __call__(self, *args, **kwargs):
        """Returns a #unrest.Rest instance. See rest entry points."""

        if self.SerializeClass is not None:
            kwargs.setdefault('SerializeClass', self.SerializeClass)
        if self.DeserializeClass is not None:
            kwargs.setdefault('DeserializeClass', self.DeserializeClass)

        rest = Rest(self, *args, **kwargs)
        return rest

    def register_options(self):
        self.framework.register_route(
            self.root_path, 'OPTIONS', None, self.unrest_api)

    def unrest_api(self):
        return self.framework.send_json(json.dumps(self.infos))

    Property = Property



from datetime import timedelta

from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import relationship
from sqlalchemy.sql.schema import Column, ForeignKey
from sqlalchemy.types import Integer, Interval, Numeric, String

Base = declarative_base()


class Tree(Base):
    """Where money doesn't grow"""
    __tablename__ = 'tree'
    id = Column(Integer, primary_key=True)
    name = Column(String)

    @property
    def fruit_colors(self):
        """The color of fruits"""
        return ', '.join([fruit.color for fruit in self.fruits])


class Fruit(Base):
    """A bag of fruit"""
    __tablename__ = 'fruit'
    fruit_id = Column('idf', Integer, primary_key=True)
    color = Column('hue', String(50))
    size = Column(Numeric)
    age = Column(Interval)
    tree_id = Column(Integer, ForeignKey('tree.id'))
    tree = relationship(Tree, backref='fruits')

    @hybrid_property
    def square_size(self):
        return self.size * self.size


def fill_data(session):
    pine = Tree(name='pine')
    maple = Tree(name='maple')
    oak = Tree(name='oak')
    session.add(pine)
    session.add(maple)
    session.add(oak)

    session.add(Fruit(
        color='grey', size=12,
        age=timedelta(days=12, hours=1, minutes=15),
        tree=pine))
    session.add(Fruit(
        color='darkgrey', size=23,
        age=timedelta(days=49, seconds=230, milliseconds=213),
        tree=pine))
    session.add(Fruit(
        color='brown', size=2.12,
        age=timedelta(0),
        tree=pine))
    session.add(Fruit(
        color='red', size=.5,
        age=timedelta(minutes=40),
        tree=maple))
    session.add(Fruit(
        color='orangered', size=100,
        age=timedelta(hours=2, microseconds=12),
        tree=maple))

    session.commit()



from . import idsorted
from ..model import Fruit, Tree


def test_get_tree_with_relationship(rest, http):
    fruit = rest(Fruit, methods=[], only=['color'])

    rest(Tree, relationships={'fruits': fruit})
    code, json = http.get('/api/tree')
    assert code == 200
    assert json['occurences'] == 3
    assert idsorted(json['objects']) == [
        {'id': 1, 'name': 'pine', 'fruits': [
            {'fruit_id': 1, 'color': 'grey'},
            {'fruit_id': 2, 'color': 'darkgrey'},
            {'fruit_id': 3, 'color': 'brown'}
        ]},
        {'id': 2, 'name': 'maple', 'fruits': [
            {'fruit_id': 4, 'color': 'red'},
            {'fruit_id': 5, 'color': 'orangered'}
        ]},
        {'id': 3, 'name': 'oak', 'fruits': []},
    ]
    code, json = http.get('/api/fruit')
    assert code == 404


def test_get_fruit_with_relationship(rest, http):
    tree = rest(Tree, methods=[])
    rest(Fruit, only=['color'], relationships={'tree': tree})

    code, json = http.get('/api/fruit')
    assert code == 200
    assert json['occurences'] == 5
    assert idsorted(json['objects'], 'fruit_id') == [{
        'fruit_id': 1,
        'color': 'grey',
        'tree': [{'id': 1, 'name': 'pine'}]
    }, {
        'fruit_id': 2,
        'color': 'darkgrey',
        'tree': [{'id': 1, 'name': 'pine'}]
    }, {
        'fruit_id': 3,
        'color': 'brown',
        'tree': [{'id': 1, 'name': 'pine'}]
    }, {
        'fruit_id': 4,
        'color': 'red',
        'tree': [{'id': 2, 'name': 'maple'}]
    }, {
        'fruit_id': 5,
        'color': 'orangered',
        'tree': [{'id': 2, 'name': 'maple'}]
    }]
    code, json = http.get('/api/tree')
    assert code == 404



import logging

from flask import current_app, jsonify, request

log = logging.getLogger('unrest.flask')


class FlaskUnRest(object):
    """
    Unrest flask framework implementation.
    This is the framework abstraction you can implement for your own framework

    """
    def __init__(self, app):
        self.app = app

    def register_route(self, path, method, parameters, fun):
        """
        Register the given function for `path` and `method` with and without
        `parameters`.

        # Arguments
            path: The url of the endoint without arguments. ('/api/person')
            method: The HTTP method to register the route on.
            parameters: The primary keys of the model that can be given
                after the path. `PrimaryKey('id'), PrimaryKey('type')) -> \
'/api/person/<id>/<type>'`
            fun: The route function
        """
        if self.app.view_functions.pop(fun.__name__, None):
            log.info('Overriding route %s' % fun.__name__)
        getattr(fun, '__func__', fun).provide_automatic_options = False
        self.app.add_url_rule(
            path, fun.__name__, fun, methods=[method])
        if parameters:
            log.info('Registering route for %s for %s' % (path, method))
            path_with_params = path + '/' + '/'.join(
                '<%s>' % param for param in parameters)

            log.info('Registering route for %s for %s' % (
                path_with_params, method))
            self.app.add_url_rule(
                path_with_params, fun.__name__, fun, methods=[method])

    def request_json(self):
        """
        Must return the string of the current JSON request content or None
        """
        if not request.is_json:
            return None

        return request.data.decode('utf-8')

    def send_json(self, json):
        """
        Send a 200 JSON response with `json` as content

        # Arguments
            json: The JSON string to send.
        """
        return current_app.response_class(
            (json, '\n'),
            mimetype=current_app.config['JSONIFY_MIMETYPE']
        )

    def send_error(self, message, status_code):
        """
        Send an error as a JSON response with the given status code.

        # Arguments
            message: The JSON string containing the error message.
            status_code: The HTTP status code (i.e. 402)
        """
        return jsonify(message), status_code
